from __future__ import print_function
#  Name   :py_code
#
#          base class for TreeDict which can
#          handle python code
#
#          py_code : add "run" to tree
#          py_script: provide script to tree
#          py_module: provide module (a set of fun)
#          py_data: container of data, can have its
#                   own module (derived from py_module)
#          py_model: container of model
#             run_all allows to run all script/modules
#             in sub tree
#
#  Inheritence
#          py_code -> py_script
#                  -> py_module -> py_data
#                  -> py_model
#
#  Example:
#
#           Params
#               - variables (data)
#           Genray-LH (model)
#               - G-file (gfile -< data )
#               - Input-file (namelist -< data)
#               - Connection (module)
#               - Solution (netcdf -< data)
#               - FigGenerator1 (script)
#               - FigGenerator2 (script)
#                        .
#                        .
#
#           Solver
#               - Loki Connection (module)
#               - Parameter Scan (script)
#               - Run Genray (module/script)
#
#
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************
from ifigure.mto.threaded_worker import ThreadedWorker
from ifigure.widgets.canvas.file_structure import *
from ifigure.mto.treedict import TreeDict
import ifigure
import os
import time
import sys
import logging
import weakref
import shutil
import wx
import ifigure.utils.pickle_wrapper as pickle
import ifigure.utils.cbook as cbook
import ifigure.events
from ifigure.utils.edit_list import DialogEditList
from ifigure.mto.hg_support import HGSupport
from ifigure.mto.treedict import TreeDict
import ifigure.widgets.dialog as dialog
from ifigure.mto.hg_support import has_repo


class PyNamespace(TreeDict):
    _save_var = False

    def __init__(self, parent=None, src=None):
        TreeDict.__init__(self, parent=parent, src=src)

    @classmethod
    def get_namebase(self):
        return 'variables'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'variable.png')
        return [idx]

    def __del__(self):
        self._var = {}

    def get_locals(self, name):
        return self.getvar(name)

    def set_locals(self, name, var):
        self.setvar(name, var)

    def new_locals(self):
        self._var = dict()
        self._var_changed = True

    def get_shownvar(self):
        names = self.get_varlist()
        hiddennames = {"obj", "proj", "top",  "param",
                       "wdir", "app", "model", "stop",
                       "exit", "write_log", "kwargs", "args", "ans"}
        for n in hiddennames:
            if n in names:
                names.remove(n)
        for n in names:
            if n.startswith('__'):
                names.remove(n)
        return names


class AbsNamespacePointer(object):
    def __init__(self):
        return self.setvar('namespace', None)

    def get_ns(self):
        return self.eval('namespace')

    def has_ns(self):
        return self.eval('namespace') is not None

    def set_ns(self, obj):
        return self.setvar('namespace', '='+obj.get_full_path())


class AbsFileContainer(object):
    def onAddNewText(self, e=None):
        from ifigure.mto.py_file import PyText
        obj = PyText()
        iname = self.get_next_name('untitled')

        parent = self.get_app()
        ret, name = dialog.textentry(parent,
                                     "Enter a text name", "New text...", iname)
        self.add_child(name, obj)
        obj.mk_owndir()
        path = os.path.join(obj.owndir(), obj.name + '.txt')
        print(('creating file ', path))
        open(path, 'a').close()
        obj.set_path_pathmode(path)
        obj.store_mtime()

        w = None if e is None else e.GetEventObject()
        from ifigure.mto.treedict import str_td
        path = str_td(path)
        path.td = obj.get_full_path()
        ifigure.events.SendEditFileEvent(self, w=w, file=path)

        if e is not None:
            e.Skip()
        return obj

    def onAddOther(self, e=None):
        from ifigure.mto.py_file import PyFile
        file = ifigure.widgets.dialog.read(None, message="Select file")
        if file == '':
            return

        dlg = wx.MessageDialog(None,
                               'Do you want to copy file to the project?\n',
                               'Opening file',
                               wx.YES_NO)
        ret = dlg.ShowModal()
        dlg.Destroy()

        if ret == wx.ID_YES:
            if not self.has_owndir():
                self.mk_owndir()
            od = self.owndir()
            new_ofile = os.path.join(od, os.path.basename(file))
            shutil.copyfile(file, new_ofile)
            file = new_ofile

        obj = PyFile()
        base, ext = obj.split_ext(os.path.basename(file))
        name = self.get_next_name(base)
        self.add_child(name, obj)
        obj.mk_owndir()
        obj.set_path_pathmode(file)
        obj.store_mtime()
        if e is not None:
            e.Skip()
        return obj

    def set_extfolderpath(self, path):
        self.setvar('ext_folder', path)

    def add_extfolder(self, basepath):

        import ifigure.mto.py_script as py_script
        from ifigure.mto.py_extfile import make_external
        if not os.path.exists(basepath):
            raise ValueError("no such directory")
        basefolder = self.add_folder(os.path.basename(basepath))
        basefolder.set_extfolderpath(basepath)
        make_external(basefolder)
        basefolder.load_extfolder()


class PyCode(TreeDict):
    def __init__(self, parent=None, src=None):
        self._break_before = False
        self._break_after = False
        self._run_done = False
        self._manual_only = False
        super(PyCode, self).__init__(parent=parent, src=src)

    @classmethod
    def isPyCode(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'code'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'script.png')
        return [idx1]

    def is_manualonly(self):
        return self._manual_only

    def run_pycode(self, e=None):
        if self._break_before:
            self._break_before = False
            return None
        # Run script
        self.do_run()

        if self._break_after:
            self._break_after = False
            return None
        return True

    def do_run(self):
        # should be overwritten
        pass

    def run_till_this(self, e=None):
        # model overwrite run so that
        # it walks throuth its sub tree and run all
        gp = self._find_grand_py_model()
        self._break_after = True
        gp.run_pycode()

    def run_before_this(self, e=None):
        gp = self._find_grand_py_model()
        self._break_before = True
        gp.run_pycode()

    def run_children(self):
        for child in self.get_children():
            if hasattr(child, 'run_pycode'):
                child.run_pycode()

    def get_root_model(self):
        def check1(obj):
            return isinstance(obj, PyModel)

        def check2(obj):
            return not isinstance(obj, PyModel)

        x = None
        for x in self.walk_tree_up(cond=check2):
            pass
        if x is not None:
            p = x.get_parent()
        else:
            p = self
        if p is None:
            return None
        if check1(p):
            for x in p.walk_tree_up(cond=check1):
                pass
            return x
        else:
            return None

    def _find_grand_py_model(self):
        parent = self.get_parent()
        try:
            if parent.PyModel():
                return parent._find_grand_py_model()
        except Exception:
            return self

#
#   Routines for TreeViewer
#
    def tree_viewer_menu(self):
        return super(PyCode, self).tree_viewer_menu()
      # return MenuString, Handler, MenuImage
      # thie following menu is suppressed until this capability
      # is necessary...
        return [('+Run...', None, None),
                ('Preceeding Steps', self.run_before_this, None),
                ('Until This Step',  self.run_till_this, None),
                ('Only This Step', self.run_pycode, None),
                ('!', None, None),
                ('---', None, None)] + \
            super(PyCode, self).tree_viewer_menu()

# save/load
    def save_data(self, fid=None):
        pass

    def load_data(self, fid=None):
        pass


class PyData(PyCode):
    '''
    PyData is the same as PyCode but different
    default icon
    '''
    @classmethod
    def isPyData(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'data'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'data.png')
        return [idx1]

    def tree_viewer_menu(self):
      # return MenuString, Handler, MenuImage
        return super(PyCode, self).tree_viewer_menu()

    def append_data(self, d):
        for k in d:
            self.setvar(k, d[k])


class AbsModuleContainer(object):
    # routines for mto which contains modules
    def onAddModule(self, e=None):
        import ifigure.mto.py_module as py_module
        child = py_module.PyModule()
        name = self.get_next_name(child.get_namebase())
        idx = self.add_child(name, child)
        if e is not None:
            e.Skip()

    def onAddNewModule(self, e=None, module_type=None):
        '''
        new module
        copy template to owndir
        open it by editor
        '''

        import ifigure.mto.py_module as py_module
        if module_type is None:
            dlg = wx.MessageDialog(self.get_app(),
                                   "Does the new module have its own directory? \n"
                                   "Or does it share the same directory as its parent? \n",
                                   "NewModule", wx.YES_NO | wx.ICON_INFORMATION)
            ans = dlg.ShowModal()
            dlg.Destroy()
            module_type = 1 if ans == wx.ID_YES else 2

        child = py_module.PyModule()
        if module_type == 1:
            temp = os.path.join(
                ifigure.__path__[0], 'template', 'new_module.py')
            child._can_have_child = True
            child._has_private_owndir = True
        else:
            temp = os.path.join(
                ifigure.__path__[0], 'template', 'new_module_nochild.py')
            child._can_have_child = False
            child._has_private_owndir = False
        name = self.get_next_name(child.get_namebase())
        idx = self.add_child(name, child)

        child.mk_owndir()
        dest = os.path.join(child.owndir(), name+'.py')
        shutil.copyfile(temp, dest)

        print(('loading...', dest))
        child.load_module(dest)

        w = None
        if e is not None:
            w = e.GetEventObject()
        from ifigure.mto.treedict import str_td
        dest = str_td(dest)
        dest.td = child.get_full_path()
        ifigure.events.SendEditFileEvent(self, w=w, file=dest)

        if e is not None:
            e.Skip()
        return child

    def add_absmodule(self, file, *args, **kargs):
        from ifigure.mto.py_module import PyModule
        child = PyModule(*args, **kargs)
        name = os.path.basename(os.path.splitext(file)[0])
        name = self.get_next_name(name)
        idx = self.add_child(name, child)
        child.load_module(file)
        return child

    def add_std_modelmodule(self, file, *args, **kargs):
        base_mod = 'ifigure.add_on.model.module'
        mod_path = cbook.GetModuleDir(base_mod)
        file = os.path.join(mod_path, file+'.py')
        return self.add_absmodule(file, *args, **kargs)

    def onAddAbsModule(self, e, file=None):
        self.add_absmodule(file)
        print(('loading add-on', file))
        e.Skip()

    def build_module_menu(self, mod_path):
        menu = []
        for file in os.listdir(mod_path):
            if (file.endswith('.py') and file.startswith('_') is False and
                    file.startswith('.') is False):
                txt = cbook.FileNameToClass(file)

                def xxx(e, file=os.path.join(mod_path, file)):
                    self.onAddAbsModule(e, file)
                menu.append((txt, xxx, None))
        return menu


class AbsScriptContainer(object):
    # routines for mto which contains script
    def onAddScript(self, e=None):
        import ifigure.mto.py_script as py_script
        child = py_script.PyScript()
        name = self.get_next_name(child.get_namebase())
        idx = self.add_child(name, child)
        if e is not None:
            e.Skip()

    def onAddNewScript(self, e=None, name='', temp='', parent=None, dest=None):
        # new script
        # copy template to owndir
        # open it by editor
        import ifigure.mto.py_script as py_script
        if temp == '':
            temp = os.path.join(
                ifigure.__path__[0], 'template', 'script', '_blank_script.py')

        child = py_script.PyScript()
        if name == '':
            iname = '_'.join([x for x in os.path.basename(temp)[
                             :-3].split('_') if len(x) != 0])
            if parent is None:
                parent = self.get_app()
            ret, name = dialog.textentry(parent,
                                         "Enter a script name", "New Script...", iname)
            if not ret:
                return
            if self.has_child(name):
                name = self.get_next_name(name)
        idx = self.add_child(name, child)

        if dest is None:
            child.mk_owndir()
            dest = child.owndir()
        dest = os.path.join(dest, name+'.py')
        i = 1
        while os.path.exists(dest):
            name0 = name+str(i)+'.py'
            dest = os.path.join(child.owndir(), name0)
            i = i + 1
        shutil.copyfile(temp, dest)
        child.load_script(dest)

        w = None
        if e is not None:
            w = e.GetEventObject()
        from ifigure.mto.treedict import str_td
        dest = str_td(dest)
        dest.td = child.get_full_path()
        ifigure.events.SendEditFileEvent(self, w=w, file=dest)

        if e is not None:
            e.Skip()
        return child

    def onAddAbsScript(self, e, file=None, name=None):
        from ifigure.mto.py_script import PyScript
        child = PyScript()
        if name is None:
            name = self.get_next_name(child.get_namebase())
        if self.has_child(name):
            print('doing here')
            name = self.get_next_name(name)
        idx = self.add_child(name, child)
        #print('loading add-on', file)
        child.load_script(file)

    def onAddScriptFromFile(self, e):

        file = dialog.read(None, message="Select script",
                           wildcard='*.py')
        if file == '':
            e.Skip()
            return
        # this call is just import check..
        from ifigure.mto.py_script import PyScript
        tmp_child = PyScript()
        idx = self.add_child('_tempraroy_script', tmp_child)
        mode, path = tmp_child.fullpath2path(file)
        if (mode == 'wdir' or mode == 'owndir'):
            m = 'Import should import from somewhere outside project directory'
            ret = dialog.message(None, message=m,
                                 title='Import error')
            return

        if not file.endswith('.py'):
            m = 'Script fils should be a .py file'
            ret = dialog.message(None, message=m,
                                 title='Import error')
            return
        name = os.path.basename(file)[:-3]

        from ifigure.widgets.dlg_fileimportmode import DlgFileimportmode

        copy, modes, copy_org = DlgFileimportmode(self,
                                                  parent=e.GetEventObject())
        if copy:
            newfile = os.path.join(self.owndir(), os.path.basename(file))
            if not self.has_owndir():
                self.mk_owndir()
            shutil.copyfile(file, newfile)
            file = newfile
            modes = ['owndir']
        mode, path = tmp_child.fullpath2path(file, modes)
        tmp_child.destroy()
        self.onAddAbsScript(e, file=file, name=name)
        e.Skip()

    def build_script_menu(self, mod_path):
        menu = []
        for file in os.listdir(mod_path):
            if (file.endswith('.py') and file.startswith('_') is False and
                    file.startswith('.') is False):
                txt = cbook.FileNameToClass(file)

                def xxx(e, file=os.path.join(mod_path, file)):
                    self.onAddAbsScript(e, file)
                menu.append((txt, xxx, None))
        return menu

    def build_script_menu2(self, mod_path):
        menu = []
        for file in os.listdir(mod_path):
            if (file.endswith('.py') and file.startswith('_') is False and
                    file.startswith('.') is False):
                txt = cbook.FileNameToClass(file)

                def xxx(e, file=os.path.join(mod_path, file)):
                    self.onAddNewScript(e, temp=file)
                menu.append((txt, xxx, None))
        return menu

    def script_template_list(self):
        base_mod = 'ifigure.template.script'
        mod_path = cbook.GetModuleDir(base_mod)
        from ifigure.ifigure_config import usr_script_template_dir

        menu = []
        menu.append(('Blank Script...', self.onAddNewScript, None))
        menu.append(('From File...', self.onAddScriptFromFile, None))
        menu.append(('+From Template...', None, None))
        menu.extend(self.build_script_menu2(mod_path))
        menu.extend(self.build_script_menu2(usr_script_template_dir))
        menu.append(('!', None, None))

        return menu


class PyModel(PyCode, AbsModuleContainer, AbsScriptContainer,
              AbsFileContainer, ThreadedWorker, HGSupport,
              AbsNamespacePointer):
    # PyModel is a top node of code tree
    # script must be empty other than
    # loading global variable to its
    # attribute
    def __init__(self, parent=None, src=None):
        self._run_script = ''
        self._gui_script = ''
        self._finish_script = ''
        self._cleanup_script = ''

        super(PyModel, self).__init__(parent=parent, src=src)
        HGSupport.__init__(self)
        ThreadedWorker.__init__(self)
        AbsNamespacePointer.__init__(self)
        if src is None:
            param = PyParam()
            self.add_child('param', param)

    @classmethod
    def isPyModel(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'model'

    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'model.png')
        idx2 = cbook.LoadImageFile(path, 'model_hg.png')
        return [idx1, idx2]

    def classimage(self):
        if PyModel._image_load_done is False:
            PyModel._image_id = self.load_classimage()
            PyModel._image_load_done = True
        if has_repo(self):
            return PyModel._image_id[1]
        else:
            return PyModel._image_id[0]

    def pv_kshortcut(self, key):
        '''
        define keyborad short cut on project viewer
        must return a method responding key
        '''
        if key == wx.WXK_F9:
            return self.onRun
        return None

    def Clean(self, event=None):
        self.do_clean()

    def do_gui(self):
        s, f, c, g = self.eval_script_location()
        if g is not None:
            g.do_run()

    def do_finish(self, *args):
        '''
        script to merge solution to PySol
        args = index, sol for parametric 
        '''
        s, f, c, g = self.eval_script_location()
        if f is not None:
            f.do_run(*args)

    def do_run(self):
        s, f, c, g = self.eval_script_location()
        if s is not None:
            if self._run_verbose:
                print(('Entering :', self.name))
                print(('   Running ', s.get_full_path()))
            s.do_run()

    def do_clean(self):
        s, f, c, g = self.eval_script_location()
        if c is not None:
            c.do_run()
        for name, treeobj in self.get_children():
            if isinstance(treeobj, PyModel):
                print(("Entering : ", name))
                treeobj.do_clean()

    def walk_model(self):
        for name, child in self.get_children():
            if isinstance(child, PyModel):
                yield child

    def tree_viewer_menu(self):
        from ifigure.mto.hg_support import has_hg
      # return MenuString, Handler, MenuImage
        s, f, c, g = self.eval_script_location()
        ss = '-' if s is None else ''
        ff = '-' if f is None else ''
        cc = '-' if c is None else ''
        gg = '-' if g is None else ''
        ll = '-' if self.has_ns() else ''

        menu = [('+Model', None,  None),
                (ss+'Run Model (F9)',   self.onRun, None),
                (gg+'Open GUI',    self.onGUI, None),
                (cc+'Run Clean',   self.onClean, None),
                ('!',  None, None),
                ('+Add...', None, None),
                ('Book',  self.onAddBook, None),
                ('Text',  self.onAddNewText, None),
                ('Data',  self.onAddData, None),
                ('Model', self.onAddModel, None),
                (ll+'Variables', self.onAddNewNamespace, None),
                ('+Script...',  None, None)] + \
            self.script_template_list() + \
            [('!',  None, None),
             ('+Module...',  None, None)] + \
            self.module_addon_list() + \
            [('!',  None, None),
                ('Other Files...', self.onAddOther, None),
                ('!',  None, None),
                ('Setting...', self.onSetting, None),
                ('---', None, None)]
        menu = self.add_hg_menu(menu)
        menu = menu + super(PyCode, self).tree_viewer_menu()
        return menu

    def onRun(self, e):
        self.Run(e)
        e.Skip()

    def onClean(self, e):
        self.Clean(e)
        e.Skip()

    def onGUI(self, e):
        self.do_gui()
        e.Skip()

    def eval_setting_str(self, txt):
        if len(txt) == 0:
            return None
        if txt[0] == '=':
            # expression is stored as it is
            try:
                root = self.get_root_parent()
                ldict = locals().copy()              
                exec(root.name + '= self.get_root_parent()', globals(),ldict)
                exec('ret'+txt, globals(),ldict)
                ret = ldict['ret']                
            except:
                import traceback
                traceback.print_exc()
                ret = None
        else:
            if txt.startswith('.'):
                txt = self.get_full_path()+txt
            try:
                root = self.get_root_parent()
                ldict = locals().copy()
                exec(root.name + '= self.get_root_parent()', globals(),ldict)
                exec('ret='+txt, globals(),ldict)
                ret = ldict['ret']
            except:
                import traceback
                traceback.print_exc()
                ret = None
        return ret

    def eval_script_location(self):
        return (self.eval_setting_str(self._run_script),
                self.eval_setting_str(self._finish_script),
                self.eval_setting_str(self._cleanup_script),
                self.eval_setting_str(self._gui_script))

    def onProjTreeActivate(self, e):
        #        pass
        self.onSetting(e)

    def onSetting(self, e):
        id = e.GetEventObject()
        app = id.GetTopLevelParent()

        r = self._run_script
        f = self._finish_script
        c = self._cleanup_script
        g = self._gui_script
        w = self._run_after

        list = [["", "Model Setting", 2],
                ["Run Script ", r, 0],
                ["GUI Script ", g, 0],
                ["Finish Script", f, 0],
                ["Cleanup Script", c, 0],
                ["Run after...", w, 0],
                [None, "(proj.model.script or .script)", 2],
                ["Threading", self._background, 3,
                 {"text": 'run this model in a separate thread'}],
                ["Run mode", self._run_mode, 3,
                 {"text": 'run a top level model script alone'}],
                ["Verbose", self._run_verbose, 3,
                 {"text": 'generate extra debug output'}]]

        app.proj_tree_viewer.OpenPanel(list, self, 'setSetting')

    def setSetting(self, value):
        self._run_script = str(value[1])
        self._gui_script = str(value[2])
        self._finish_script = str(value[3])
        self._cleanup_script = str(value[4])
        self._run_after = str(value[5])
        self._background = value[7]
        self._run_mode = value[8]
        self._run_verbose = value[9]

    def set_guiscript(self, value):
        self._gui_script = value

    def set_runscript(self, value):
        self._run_script = value

    def set_finishscript(self, value):
        self._finish_script = value

    def set_cleanupscript(self, value):
        self._cleanup_script = value

    def get_guiscript(self):
        return self.eval_setting_str(self._gui_script)

    def get_runscript(self):
        return self.eval_setting_str(self._run_script)

    def get_finishscript(self):
        return self.eval_setting_str(self._finish_script)

    def get_cleanupscript(self):
        return self.eval_setting_str(self._cleanup_script)

    def onAddBook(self, e):
        from ifigure.mto.fig_book import FigBook
        book = FigBook()
        name = self.get_next_name('book')
        self.add_child(name, book)
        e.Skip()  # call update project tree widget

    def onAddNewNamespace(self, e=None):
        ns = PyNamespace()
        if self.has_child('variables'):
            name = self.get_next_name('variables')
        else:
            name = 'variables'
        self.add_child(name, ns)
        self.set_ns(ns)
        if e is not None:
            e.Skip()  # call update project tree widget

    def onAddData(self, e):
        id = e.GetEventObject()
        app = id.GetTopLevelParent()
        ret, name = dialog.textentry(app,
                                     "Enter a data name",
                                     "New Data...", "data")
        if not ret:
            return
        data = PyData()
        name = self.get_next_name(name)
        self.add_child(name, data)
        e.Skip()  # call update project tree widget

    def onAddModel(self, e):
        id = e.GetEventObject()
        app = id.GetTopLevelParent()
        ret, new_name = dialog.textentry(app,
                                         "Enter the name of new model",
                                         "Add Model", "")
        if ret:
            child = PyModel()
            imodel = self.add_child(new_name, child)
            child.onAddFolder(name='scripts')
        e.Skip()  # call update project tree widget

    def onAddExtFolder(self, evt):
        from ifigure.widgets.dialog import readdir
        dir = readdir(parent=evt.GetEventObject(),
                      message='Select external folder to import')
        if dir == '':
            return
        self.add_extfolder(dir)
        if evt is not None:
            evt.Skip()

    def add_folder(self, name):
        return self.add_childobject(PyFolder, name)

    def add_book(self, name):
        from ifigure.mto.fig_book import FigBook
        return self.add_childobject(FigBook, name)

    def add_data(self, name):
        return self.add_childobject(PyData, name)

    def add_model(self, name, cls=None):
        if cls is None:
            cls = PyModel
        return self.add_childobject(cls, name)

    def add_file(self, name):
        from ifigure.mto.py_file import PyFile
        return self.add_childobject(PyFile, name)

    def add_text(self, name):
        from ifigure.mto.py_file import PyText
        return self.add_childobject(PyText, name)

    def module_addon_list(self):
        base_mod = 'ifigure.add_on.model.module'
        mod_path = cbook.GetModuleDir(base_mod)
        menu = self.build_module_menu(mod_path)
        menu.append(('New Module...', self.onAddNewModule, None))
        return menu

    def script_addon_list(self):
        base_mod = 'ifigure.add_on.model.script'
        mod_path = cbook.GetModuleDir(base_mod)
        menu = self.build_script_menu(mod_path)
        menu.append(('New Script...', self.onAddNewScript, None))
        return menu

    def save_data2(self, data=None):
        if data is None:
            data = {}
        h2 = {"r": self._run_script,
              "f": self._finish_script,
              "c": self._cleanup_script,
              "g": self._gui_script,
              "background": self._background,
              "run_after": self._run_after,
              "run_mode": self._run_mode,
              "run_verbose": self._run_verbose}
        data['PyModel'] = (1, h2)
        return super(PyModel, self).save_data2(data)

    def load_data2(self, data):
        h2 = data['PyModel'][1]
        if "r" in h2:
            self._run_script = h2["r"]
        if "c" in h2:
            self._cleanup_script = h2["c"]
        if "f" in h2:
            self._finish_script = h2["f"]
        if "g" in h2:
            self._gui_script = h2["g"]
        if "background" in h2:
            self._background = h2["background"]
        if "run_after" in h2:
            self._run_after = str(h2["run_after"])
        if "run_mode" in h2:
            self._run_mode = h2["run_mode"]
        if "run_verbose" in h2:
            self._run_verbose = h2["run_verbose"]
        return super(PyModel, self).load_data2(data)


class PyParam(PyCode):
    '''
    PyParams is a object which provide an convenient way to
    specify a parameter sets. Each model can have one PyParam
    and evaluation of PyParam can be overwritten.


    '''

    def __init__(self, parent=None, src=None):
        super(PyParam, self).__init__(parent=None, src=None)
        self._name_readonly = True
        self._manual_only = True
        self._local_vars = []

    @classmethod
    def isPyParam(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'param'

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'pi.png')
        return [idx1]

    def run_pycode(self):
        # PyParam has nothing to run
        pass

    def tree_viewer_menu(self):
      # return MenuString, Handler, MenuImage
        list = super(PyCode, self).tree_viewer_menu()
        for i in range(len(list)):
            if list[i][0] == 'Delete':
                list[i] = ('-Delete', None, None)
                break
        for i in range(len(list)):
            if list[i][0] == 'Suppress':
                list[i] = ('-Suppress', None, None)
                break
        for i in range(len(list)):
            if list[i][0] == 'Unsuppress':
                list[i] = ('-Unsuppress', None, None)
                break

        return [('Print All Variables', self.onPrint, None),
                ('Export to Main', self.onExport, None),
                ('Import from Main', self.onImport, None),
                ('---', None, None)] + \
            list

    def eval_all_keys(self, path=False):
        d = {}  # dict to make unique set....
        if self.get_parent() is None:
            '''
            safty if paranet is None
            '''
            if path:
                return d
            else:
                return list(d.keys())

        for item in self.get_parent().walk_tree_up():
            for name, child in item.get_children():
                if isinstance(child, PyParam):
                    for key in child.getvar():
                        if key not in d:
                            d[key] = [(child.get_full_path(), child)]
                        else:
                            d[key].append((child.get_full_path(), child))
                    break
        try:
            item = self.get_root_parent().setting.parameters
            if item is self:
                if path:
                    return d
                else:
                    return list(d.keys())

        except:
            if path:
                return d
            else:
                return list(d.keys())

        for key in item.getvar():
            if key not in d:
                d[key] = (item.get_full_path(), item)
            else:
                d[key].append((item.get_full_path(), item))
        if path:
            return d
        return list(d.keys())

    def set(self, name, var):
        d = self.eval_all_keys(path=True)
        if not name in d:
            target = self
        else:
            for x in d[name]:
                if name in x[1]._local_vars:
                    break
            target = x[1]
#            target = d[name][-1][1]
        TreeDict.setvar(target, name,  var)

    def get(self, name):
        return self.eval(name)

    def eval(self, name):
        '''
        params._var can be overwritten by upper level
        '''
        if not name in self._var:
            print(('Warning : PyParam::get: Variable is not set at this PyParam,',
                   name, ':', self.get_full_path()))

        d = self.eval_all_keys(path=True)
        if not name in d:
            return None
        for path, obj in d[name]:
            if name in obj._local_vars:
                break
        val = TreeDict.eval(obj, name)
        return val

    def onPrint(self, e):
        d = self.eval_all_keys(path=True)
        for key in d:
            print((key, 'defined as...'))
            for x in d[key]:
                val = x[1]._var[key]
                if isinstance(val, str):
                    if val.startswith('='):
                        val = eval(val[1:])

                print(('   ', str(val), ' at ',  x[1].get_full_path()))

    def onExport(self, e):
        d = {}
        for key in self.eval_all_keys():
            d[key] = self.eval(key)
        list = [["", "Enter Variable Name to Export", 2],
                ["Name", "var", 0]]
        flag, value = DialogEditList(list)
        if flag:
            self.write2shell(d,  value[1])

    def onImport(self, e):
        list = [["", "Enter Variable Name to Import", 2],
                ["Name", "var", 0]]
        flag, value = DialogEditList(list)
        if flag:
            var = cbook.ReadFromMain(value[1])
            for key in var:
                self.setvar(key, var[key])

    def onDelete(self, e=None):
        # prohibit delete pyparam interactively
        pass

    def save_data2(self, data=None):
        if data is None:
            data = {}
        h2 = {"_local_vars": self._local_vars, }
        data['PyParam'] = (1, h2)
        return super(PyParam, self).save_data2(data)

    def load_data2(self, data):
        if not 'PyParam' in data:
            return
        h2 = data['PyParam'][1]
        for key in h2:
            setattr(self, key, h2[key])
        return super(PyParam, self).load_data2(data)

    def set_var_local(self, name, flag=True):
        if flag:
            if not name in self._local_vars:
                self._local_vars.append(name)
        else:
            if name in self._local_vars:
                del self._local_vars[name]

    def delvar(self, name):
        PyCode.delvar(self, name)
        if name in self._local_vars:
            self._local_vars.remove(name)


class PyFolder(TreeDict, AbsModuleContainer, AbsScriptContainer,
               AbsFileContainer, HGSupport):
    def __init__(self, *args, **kargs):
        super(PyFolder, self).__init__(*args, **kargs)
        HGSupport.__init__(self)

    @classmethod
    def isPyFolder(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'folder'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'folder.png')
        idx2 = cbook.LoadImageFile(path, 'folder_hg.png')
        idx3 = cbook.LoadImageFile(path, 'folder_ext.png')
        return [idx1, idx2, idx3]

    def classimage(self):
        if PyFolder._image_load_done is False:
            PyFolder._image_id = self.load_classimage()
            PyFolder._image_load_done = True
        if has_repo(self):
            return PyFolder._image_id[1]
        elif self.getvar('ext_folder') is not None:
            return PyFolder._image_id[2]
        else:
            return PyFolder._image_id[0]

    def tree_viewer_menu(self):
        from ifigure.mto.hg_support import has_hg
        # return MenuString, Handler, MenuImage
        menu = [
            ('Add Text',   self.onAddNewText, None),
            ('+Add Script...',  None, None)] + \
            self.script_template_list() + \
            [('!',  None, None),
             ('Other File...',   self.onAddOther, None),
             ('---', None, None)]
        menu = self.add_hg_menu(menu)
        menu = menu + super(PyFolder, self).tree_viewer_menu()
        return menu

    def add_folder(self, name):
        return self.add_childobject(PyFolder, name)

    def add_book(self, name):
        from ifigure.mto.fig_book import FigBook
        return self.add_childobject(FigBook, name)

    def add_data(self, name):
        return self.add_childobject(PyData, name)

    def add_file(self, name):
        from ifigure.mto.py_file import PyFile
        return self.add_childobject(PyFile, name)

    def add_text(self, name):
        from ifigure.mto.py_file import PyText
        return self.add_childobject(PyText, name)

    def onAddNewScript(self, *args, **kwargs):
        if self.get_extfolderpath() is None:
            super(PyFolder, self).onAddNewScript(*args, **kwargs)
        else:
            kwargs['dest'] = self.get_extfolderpath()
            super(PyFolder, self).onAddNewScript(*args, **kwargs)

    def onAddExtFolder(self, evt):
        from ifigure.widgets.dialog import readdir
        dir = readdir(parent=evt.GetEventObject(),
                      message='Select library directory to import')
        if dir == '':
            return
        self.add_extfolder(dir)
        if evt is not None:
            evt.Skip()

    def load_script_folder(self, basepath, skip_underscore=False):
        from ifigure.mto.py_script import PyScript
        warning = []
        for dirpath, dirname, filenames in os.walk(basepath):
            if '.' in os.path.basename(dirpath):
                continue
            dirname = [x for x in dirname if not '.'in x]
            relpath = os.path.relpath(dirpath, basepath)
            if relpath == '.':
                relpath = ''
            p = self

            tmp = relpath.split(os.sep)
            if any(['.' in x for x in tmp]):
                continue
            for name in relpath.split(os.sep):
                if name == '':
                    continue
                p = p.get_child(name=name)

            for name in dirname:
                obj = PyFolder()
                p.add_child(name, obj, warning=warning)
            for f in filenames:
                fpath = os.path.join(dirpath, f)

                if fpath.endswith('.py'):
                    newname = str(os.path.basename(fpath).split('.')[0])
                    if skip_underscore and newname.startswith('_'):
                        continue
                    child = PyScript()
                    idx = p.add_child(newname, child, warning=warning)
                    child.import_script(fpath)
#                    script.onImportScriptFile(file = fpath)
        for x in warning:
            if len(x) != 0:
                print(x)


class PySol(TreeDict):
    """
    PySol is a folder to store solution
    """

    def __init__(self, *args, **kargs):
        self._lock = None
        TreeDict.__init__(self, *args, **kargs)

    @classmethod
    def isPySol(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'solution'

    @classmethod
    def can_have_child(self, child=None):
        return True

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'data.png')
        return [idx1]

    def tree_viewer_menu(self):
        has_pyparam = False
        for name, child in self.get_children():
            if isinstance(child, PyParam):
                has_pyparam = True
                break

        # return MenuString, Handler, MenuImage
        if has_pyparam:
            return [('-Add Parameters', self.onAddParam, None), ] + \
                super(PySol, self).tree_viewer_menu()
        else:
            return [('Add Parameters', self.onAddParam, None), ] + \
                super(PySol, self).tree_viewer_menu()

    def onAddParam(self, evt):
        obj = PyParam()
        name = 'parameters'
        self.add_child(name, obj)
        evt.Skip()  # call update project tree widget

    def aquire_lock(self, *args, **kargs):
        import threading
        if self._lock is None:
            self._lock = threading.Lock()
        self._lock.acquire(*args, **kargs)

    def release_lock(self):
        self._lock.release()



#!/usr/bin/python

from __future__ import print_function
import socket
import threading
import os
import signal
import sys
import time
import logging
import traceback
from six.moves import socketserver
#import paramiko
import binascii
from six.move import cPickle as pickle
#import cPickle as pickle
from ifigure.utils.daemon import Daemon
import MDSplus
from MDSplus import Connection
#from .utils import parse_server_string

write_handler_log = False


def parse_server_string(txt):
    l = txt.split(':')
    p = l[-2]
    t = l[-1]
    s = ':'.join(l[:-2])
    return s, p, t


class ForkingTCPRequestHandler(socketserver.StreamRequestHandler):
    #
    def __init__(self, *args, **kargs):
        name = '/tmp/mdsplus_gateway_'+str(os.getpid()) + '.log'

        self.use_mdsconnect = False
        self.connection_str = ''
        self.connection = None
        if write_handler_log:
            self.fid = open(name, 'w')
        else:
            self.fid = None
        socketserver.StreamRequestHandler.__init__(self, *args, **kargs)

    def write_log(self, txt):
        if self.fid is not None:
            self.fid.write(txt + '\n')

    def handle(self):
        try:
            while True:
                self.loop_handler()
        except:
            pass

    def loop_handler(self):
        message = self.rfile.readline().strip()
        if self.fid is not None:
            self.fid.write('handling input \n' + message + '\n')
        asmessage = binascii.a2b_hex(message)
#        logging.basicConfig(level = logging.DEBUG)
        try:
            data = pickle.loads(asmessage)
        except ValueError as EOFError:
            if self.fid is not None:
                self.fid.write('picke error \n')
                self.fid.flush()
#           logging.exception('picke load error', len(message))
#           response = None
#           return
        #cur_thread = threading.current_thread()
        if self.fid is not None:
            self.fid.write('handling request  \n' + str(data) + '\n')
            self.fid.flush()
        r = {}
        for name, commands in data:
            for command in commands:
                com = command[0]
                param = command[1:]
        #cur_thread = threading.current_thread()
                if self.fid is not None:
                    self.fid.write(com + ':' + param + '\n')
                    self.fid.flush()
                if com == 's':

                    server, port, tree = parse_server_string(param)
                    self.use_mdsconnect = (False if server.upper() ==
                                           'DIRECT' else True)

                    if self.use_mdsconnect:
                        try:
                            if self.connection_str != param:
                                del self.connection
                                if port != '':
                                    server = server + ':' + port
                                self.connection = Connection(server)
                                self.connection_str = param
                                r[name] = 'ok'
                        except:
                            r[name] = 'connection error'
                            self.error = ['connection error',
                                          self.connection_str]
                    else:
                        if self.connection is not None:
                            del self.connection
                            self.connection = None
                        self.connection_str = ''
                        r[name] = 'ok'
                else:
                    self.error = []
                    if self.use_mdsconnect:
                        r[name] = self._handle_mdsconnect(com, param)
                    else:
                        r[name] = self._handle(com, param)
                    if len(self.error) != 0:
                        if self.fid is not None:
                            self.fid.write(str(self.error))
                            self.fid.flush()
        if self.fid is not None:
            self.fid.write('return variables  \n' + str(list(r.keys())) + '\n')
            self.fid.flush()

        sr = pickle.dumps(r)
        response = binascii.b2a_hex(sr)
        if self.fid is not None:
            self.fid.write('sending data (length) \n' +
                           str(len(response)) + '\n')
            self.fid.flush()
        self.request.sendall(response+'\n')
        if self.fid is not None:
            self.fid.flush()

    def _handle_mdsconnect(self, com, param):
        response = ''

        self.write_log(str(com)+':'+str(param))
        if com == 'c':  # connection request
            arr = param.split(',')
            tree = ','.join(arr[:-1])
            shot = arr[-1]
            try:
                if tree != '':
                    self.write_log('opening tree :' +
                                   str(tree) + ':'+str(shot))
                    self.connection.openTree(tree, int(shot))
                response = 'ok'
            except Exception:
                self.error = ['run error', traceback.format_exc()]
                response = None
        elif com == 'd':
            node = param
            self.write_log('setting def node ' + str(node))
            try:
                if node.strip() != '':
                    self.connection.setDefault(node)
                response = 'ok'
            except Exception:
                self.error = ['run error', traceback.format_exc()]
                response = None
        elif com == 'v':  # mdsvalue
            self.write_log('mdsvalue '+str(param))
            try:
                response = self.connection.get(param).data()
            except Exception:
                self.error = ['run error', param, traceback.format_exc()]
                response = None
        elif com == 'u':  # mdsvalue
            self.write_log('mdsvalue '+str(param))
            try:
                response = self.connection.get(
                    '_piscopevar=execute($)', param).data()
            except Exception:
                self.error = ['run error', param, traceback.format_exc()]
                response = None
        elif com == 'w':  # dim_of (this is not used anymore)
            self.write_log('dim_of '+str(param))
            expr = param[1:]
            dim = int(param[0])
            try:
                #               response =MDSplus.Data.compile(param).evaluate().data()
                response = self.connection.get(expr).dim_of(dim).data()
            except Exception:
                self.error = ['run error', param, traceback.format_exc()]
                response = None
        # self.write_log(response.__repr__())
        return response

    def _handle(self, com, param):
        response = ''

        self.write_log(str(com)+' '+str(param))
        if com == 'c':  # connection request
            arr = param.split(',')
            tree = ','.join(arr[:-1])
            shot = arr[-1]
            try:
                if tree != '':
                    self.write_log('opening tree '+str(tree) + ' '+str(shot))
                    self.t = MDSplus.Tree(tree, int(shot))
                    self.write_log('opening tree '+str(self.t))
                response = 'ok'
            except Exception:
                self.error = ['run error', traceback.format_exc()]
                response = None
        elif com == 'd':
            node = param
            self.write_log('setting def node '+str(node))
            try:
                if node.strip() != '':
                    tn = self.t.getNode(node)
                    self.t.setDefault(tn)
                response = 'ok'
            except Exception:
                self.error = ['run error', traceback.format_exc()]
                response = None
        elif com == 'f':  # connection request
            a = param.split(',')
            tree = a[0]
            shot = a[1]
            node = a[2]
            expr = ','.join(a[3:])
            try:
                self.t = MDSplus.Tree(tree, int(shot))
                if node.strip() != '':
                    tn = self.t.getNode(node)
                    self.t.setDefault(tn)
                response = MDSplus.Data.compile(expr).evaluate().data()
            except Exception:
                self.error = ['run error', expr, traceback.format_exc()]
                response = None
        elif com == 'v':  # mdsvalue
            self.write_log('mdsvalue '+str(param))
            try:
                response = MDSplus.Data.compile(param).evaluate().data()
            except Exception:
                self.error = ['run error', param, traceback.format_exc()]
                response = None
        elif com == 'u':  # mdsvalue
            self.write_log('mdsvalue '+str(param))
            try:
                response = MDSplus.Data.execute(
                    '_piscopevar=execute($)', param).data()
            except Exception:
                self.error = ['run error', param, traceback.format_exc()]
                response = None
        elif com == 'w':  # dim_of (this is not used anymore)
            self.write_log('dim_of '+str(param))
            expr = param[1:]
            dim = int(param[0])
            try:
                #               response =MDSplus.Data.compile(param).evaluate().data()
                response = MDSplus.Data.execute(expr).dim_of(dim).data()
            except Exception:
                self.error = ['run error', param, traceback.format_exc()]
                response = None
        return response


class ForkingTCPServer(socketserver.ForkingMixIn, socketserver.TCPServer):
    pass


class MDSDaemon(Daemon):
    def __init__(self, HOST, PORT, pidfile, stdout='/dev/null',
                 stderr='/dev/null'):
        self.host = HOST
        self.port = PORT
        Daemon.__init__(self, pidfile, stdout=stdout, stderr=stderr)

    def run(self):
        self.server = ForkingTCPServer((self.host, self.port),
                                       ForkingTCPRequestHandler)

        print(('Starting MDSplus Gateway', self.host, self.port))
        print(sys.stdout)
        self.ip, self.port = self.server.server_address
        sys.stdout.flush()
        sys.stderr.flush()

        # Start a thread with the server -- that thread will then start one
        # more thread for each request
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        # Exit the server thread when the main thread terminates
        self.server_thread.setDaemon(True)
        self.server_thread.start()
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

        while True:
            # this loop is to keep thread on
            time.sleep(0.1)
            sys.stdout.flush()
            sys.stderr.flush()

    def signal_handler(self, signal, frame):
        print(('Received singal', signal))
        self.server.shutdown()
        sys.exit(0)

    def stop(self):
        print('Stopping server')
        Daemon.stop(self)


if __name__ == "__main__":
    HOST, PORT = socket.gethostname(), 10002
    pid_file = '/tmp/mdsplus_gateway.pid'
    log_file = '/tmp/mdsplus_gateway.log'

    if len(sys.argv) == 2:
        if 'start' == sys.argv[1]:
            if os.path.exists(log_file):
                os.remove(log_file)
            daemon = MDSDaemon(HOST, PORT, pid_file,
                               stdout=log_file,
                               stderr=log_file)
            daemon.start()
        elif 'stop' == sys.argv[1]:
            daemon = MDSDaemon(HOST, PORT, pid_file,
                               stdout=log_file,
                               stderr=log_file)
            daemon.stop()
        elif 'restart' == sys.argv[1]:
            daemon = MDSDaemon(HOST, PORT, pid_file,
                               stdout=log_file,
                               stderr=log_file)
            daemon.restart()
        else:
            print("Unknown command")
            sys.exit(2)
            sys.exit(0)
    else:
        print("usage: %s start|stop|restart" % sys.argv[0])
        sys.exit(2)



import ifigure.utils.debug as debug
from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.axis_user import XUser, YUser, ZUser, CUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
from ifigure.utils.cbook import ProcessKeywords, LoadImageFile, isdynamic
from matplotlib.tri import Triangulation
from ifigure.matplotlib_mod.triplot_mod import triplot
from ifigure.utils.args_parser import ArgsParser
from matplotlib.colors import Normalize, colorConverter, LightSource, Colormap
from matplotlib.cm import ScalarMappable
from matplotlib.colors import ColorConverter

from ifigure.mto.axis_param import get_cmap

cc = ColorConverter()


dprint1, dprint2, dprint3 = debug.init_dprints('FigSurface')


def _reduce_size(X, Y, Z, rstride, cstride):
    rows, cols = Z.shape
    # Force X and Y to take the same shape.
    # If they can not be fitted to that shape,
    # then an exception is automatically thrown.
    X.shape = (rows, cols)
    Y.shape = (rows, cols)

    # We want two sets of lines, one running along the "rows" of
    # Z and another set of lines running along the "columns" of Z.
    # This transpose will make it easy to obtain the columns.
    rii = range(0, rows, rstride)
    cii = range(0, cols, cstride)

    # Add the last index only if needed
    if rows > 0 and rii[-1] != (rows - 1):
        rii += [rows-1]
    if cols > 0 and cii[-1] != (cols - 1):
        cii += [cols-1]
    return X[rii][:, cii],  Y[rii][:, cii], Z[rii][:, cii]

#from mpl_toolkits.axes_grid1.inset_locator import inset_axes


class FigSurface(FigObj, XUser, YUser, ZUser, CUser):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._objs = []  # for debug....
            obj._data_extent = None
            obj._coarse_artist = None
            obj._fine_artist = None
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_opt('x', None, ['iter|nonstr', 'dynamic'])
        p.add_opt('y', None, ['iter|nonstr', 'dynamic'])
        p.add_var('z', ['iter|nonstr', 'dynamic'])

        p.add_key('cmap', None)
        p.add_key('shade', 'flat')
        p.add_key('linewidth', 1.0)
        p.add_key('alpha', None)
        p.add_key('rstride', 1)
        p.add_key('cstride', 1)
        p.add_key('edgecolor', 'k')
        p.add_key('facecolor', 'b')
        p.add_key('cz', not 'color' in kywds)
        p.add_key('cdata', None)

        p.set_pair("x", "y")
        p.set_ndconvert("x", "y", "z")
        p.set_squeeze_minimum_1D("x", "y", "z")

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)

        obj = set_hidden_vars(obj)
        if v['cmap'] is not None:
            if isinstance(v['cmap'], Colormap):
                v['cmap'] = v['cmap'].name
            kywds['cmap'] = v['cmap']
            del v['cmap']

        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    def __init__(self, *args,  **kywds):
        XUser.__init__(self)
        YUser.__init__(self)
        ZUser.__init__(self)
        CUser.__init__(self)

        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigSurface, self).__init__(*args, **kywds)
        self._method = 'plot_surface'

    @classmethod
    def get_namebase(self):
        return 'surface'

    @classmethod
    def property_in_palette(self):
        return ["facecolor_2", "edgecolor_2", "linewidthz", "solid_shade",
                "alpha_2", "noclip3d"]

    @classmethod
    def attr_in_file(self):
        return (["cmap", "cstride", "rstride", "shade", "alpha",
                 "edgecolor", "linewidth", "facecolor"] +
                super(FigSurface, self).attr_in_file())

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'surf.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        ZUser.unset_az(self)
        CUser.unset_ac(self)

        super(FigSurface, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)
        ZUser.get_zaxisparam(self)
        CUser.get_caxisparam(self)

    def args2var(self):
        ret = self._args2var()
        if ret:
            if 'cmap' in self.getvar("kywds"):
                cax = self.get_caxisparam()
                cax.set_cmap(self.getvar('kywds')['cmap'])
        return ret

    def _args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y", "z"]
        use_np = [True]*3
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)
        x = values[0]
        y = values[1]
        z = values[2]
        if x is None and y is None:
            xtmp = np.arange(z.shape[1])
            ytmp = np.arange(z.shape[0])
            x, y = np.meshgrid(xtmp, ytmp)
            self.setp("x", x)
            self.setp("y", y)
        if (x.ndim == 1 and
                y.ndim == 1):
            x, y = np.meshgrid(x, y)
            self.setp("x", x)
            self.setp("y", y)
        if (x.ndim != 2 or
            y.ndim != 2 or
                z.ndim != 2):
            return False

        if x is None:
            return False
        if y is None:
            return False
        if z is None:
            return False

        return True

    def generate_artist(self, coarse=False):
        if (self.isempty() is False
                and not coarse):
            return

        axes = self.get_figaxes()
        if not axes.get_3d():
            return

        container = self.get_container()
        x, y, z = self._eval_xy()
        # use_var should be false if evaluation is
        # okey.
        if self.getp('use_var'):
            return

        kywds = self._var["kywds"].copy()

        cax = self.get_caxisparam()
        if cax is None:
            dprint1('Error: cax is None')
            return

        kywds['alpha'] = self.getp('alpha')
        ### stride is handled in fig_surface
        kywds['cstride'] = self.getvar('cstride')
        kywds['rstride'] = self.getvar('rstride')
        if coarse:
            f = ifigure._visual_config["3d_rot_accel_factor"]
        else:
            f = 1
        x, y, z = _reduce_size(x, y, z,
                               self.getp('cstride')*f,
                               self.getp('rstride')*f)

        if (not 'color' in kywds and
                not 'facecolors' in kywds):
            cmap = self.get_cmap()
            kywds['cmap'] = get_cmap(cmap)

        # if self.getp('alpha') is not None else 1
        kywds['alpha'] = self.getp('alpha')
        fc = self.getp('facecolor')
        if isinstance(fc, str):
            fc = cc.to_rgba(fc)
        if fc is None:
            fc = [0, 0, 0, 0]
        else:
            fc = list(fc)
            if self.getp('alpha') is not None:
                fc[3] = self.getp('alpha')
        ec = self.getp('edgecolor')
        if isinstance(ec, str):
            ec = cc.to_rgba(ec)
        if ec is None:
            ec = [0, 0, 0, 0]
        else:
            ec = list(ec)
            if self.getp('alpha') is not None:
                ec[3] = self.getp('alpha')
        cz = self.getvar('cz')
        if cz is None:
            kywds['cz'] = False
        else:
            kywds['cz'] = cz
        if kywds['cz']:
            kywds['cdata'] = self.getvar('cdata')
        kywds['facecolor'] = (fc,)
        kywds['edgecolor'] = (ec,)
        kywds['linewidths'] = 0.0 if self.getp(
            'linewidth') is None else self.getp('linewidth')
        kywds['shade'] = self.getvar('shade')

        m = getattr(container, self._method)
        self._artists = [m(x, y, z, **kywds)]
        self._fine_artist = self._artists[0]

        for artist in self._artists:
            artist.do_stencil_test = False
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))
            self._objs.append(weakref.ref(artist))
            cax.set_crangeparam_to_artist(artist)

    def switch_scale(self, level='fine'):
        return
        col = self.get_figaxes()._artists[0].collections
        if self._coarse_artist is None:
            self._coarse_artist = self.generate_artist(coarse=True)
            self._coarse_artist.figobj = self

        dprint1('collection ', col)
        if level == 'fine':
            if self._coarse_artist in col:
                col.remove(self._coarse_artist)
            if not self._fine_artist in col:
                col.append(self._fine_artist)
            self._artists[0] = self._fine_artist
        elif level == 'coarse':
            if self._fine_artist in col:
                col.remove(self._fine_artist)
            if not self._coarse_artist in col:
                col.append(self._coarse_artist)
            self._artists[0] = self._coarse_artist

    def del_artist(self, artist=None, delall=False):
        if (len(self._artists) == 1 and
                self._artists[0] != self._fine_artist):
            self.switch_scale(level='fine')

        artistlist = self._artists
        self.store_loaded_property()

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                # GL canvas check this if artist is still alive.
                a.axes = None
                a.remove()

        if self._coarse_artist is not None:
            col = self.get_figaxes()._artists[0].collections
            if self._coarse_artist in col:
                self._coarse_artist.remove()
            self._coarse_artist.figobj = None
            self._coarse_artist = None
        self._fine_artist = None
        super(FigSurface, self).del_artist(artistlist)

    def get_mappable(self):
        return [a for a in self._artists if isinstance(a, ScalarMappable)]

    def highlight_artist(self, val, artist=None):
        from ifigure.matplotlib_mod.art3d_gl import Poly3DCollectionGL
        figure = self.get_figpage()._artists[0]
        ax = self.get_figaxes()
        if artist is None:
            alist = self._artists
        else:
            alist = artist

        if val == True:
            if self._parent is None:
                return
            container = self.get_container()
            if container is None:
                return

            de = self.get_data_extent()
            x = (de[0], de[1], de[1], de[0], de[0])
            y = (de[2], de[2], de[3], de[3], de[2])

            facecolor = 'k'
            if isinstance(alist[0], Poly3DCollectionGL):
                hl = alist[0].add_hl_mask()
#               hl = alist[0].make_hl_artist(container) # slower
                facecolor = 'none'
            else:
                hl = []

            for item in hl:
                alist[0].figobj_hl.append(item)

        else:
            for a in alist:
                if len(a.figobj_hl) == 0:
                    continue
                for hl in a.figobj_hl:
                    hl.remove()
                a.figobj_hl = []

#
#   Setter/Getter
#
    def set_cmap(self, value, a):
        a.set_cmap(get_cmap(value))
        self.setp('cmap', value)
        ca = self.get_caxisparam()
        ca.set_cmap(value)
        if self.has_cbar():
            ca.update_cb()

    def get_cmap(self, a=None):
        ca = self.get_caxisparam()
        return ca.cmap

    def set_alpha(self, value, a):
        a.set_alpha(value)
        self.setp('alpha', value)
        #self.setp('cmap', value)
        self.set_cmap(self.get_cmap(a), a)

    def get_alpha(self, a=None):
        return self.getp('alpha')

    def set_edgecolor(self, value, a):
        self.setp('edgecolor', value)
        a.set_edgecolor([value])

    def get_edgecolor(self, a=None):
        return self.getp('edgecolor')

    def set_linewidth(self, value, a):
        self.setp('linewidth', value)
        a.set_linewidth(value)

    def get_linewidth(self, a=None):
        return self.getp('linewidth')

    def set_shade(self, value, a):
        self.setvar('shade', value)
        self.del_artist(delall=True)
        self.delp('loaded_property')

        self.generate_artist()
        ax = self.get_figaxes()
        ax.set_bmp_update(False)

        sel = [weakref.ref(self._artists[0])]
        import wx
        app = wx.GetApp().TopWindow
        # ifigure.events.SendPVDrawRequest(self, w=None,
        #                        wait_idle=True, refresh_hl=True)
        ifigure.events.SendSelectionEvent(self, w=app, selections=sel)
        # self.reset_artist()

    def get_shade(self, a=None):
        return self.getvar('shade')

    def set_facecolor(self, value, a):
        if self.getvar('cz'):
            return
        if value == 'disabled':
            return
        if isinstance(value, str):
            value = cc.to_rgba(value)
        alpha = self.getp('alpha')
        if alpha is None:
            alpha = 1.0
        value = (value[0], value[1], value[2], alpha)
        self.setp('facecolor', value)
        a.set_facecolor([value])

    def get_facecolor(self, a=None):
        if self.getvar('cz'):
            return 'disabled'
        return self.getp('facecolor')

#
#   def hit_test
#
    def picker_a(self, artist, evt):
        axes = artist.axes
        if axes is None:
            return False, {}
        hit, extra = artist.contains(evt)

        if hit:
            return True, {}
        else:
            return False, {}

    def get_data_extent(self):
        if self._data_extent is not None:
            return self._data_extent
        x, y, z = self._eval_xy()
        self._data_extent = [np.min(x), np.max(x),
                             np.min(y), np.max(y),
                             np.min(z), np.max(z)]
        return self._data_extent

    def _update_range_gl(self, range, idx):
        for a in self._artists:
            if hasattr(a, 'get_gl_data_extent'):
                tmprange = a.get_gl_data_extent()[idx]
            range = self._update_range(range,
                                       tmprange)
        return range

    def get_xrange(self, xrange=[None, None], scale='linear'):
        x, y, z = self._eval_xy()
        if x is None:
            return xrange
        if scale == 'log':
            x = mask_negative(x)
        return self._update_range(xrange, (np.nanmin(x), np.nanmax(x)))

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        x, y, z = self._eval_xy()
        if x is None:
            return yrange
        if y is None:
            return yrange
        if scale == 'log':
            y = mask_negative(y)
        return self._update_range(yrange, (np.nanmin(y), np.nanmax(y)))

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):

        x, y, z = self._eval_xy()
        if (xrange[0] is not None and
            xrange[1] is not None and
            yrange[0] is not None and
                yrange[1] is not None):
            if x.size*y.size == z.size:
                idx1 = np.where((y >= yrange[0]) & (y <= yrange[1]))[0]
                idx2 = np.where((x >= xrange[0]) & (x <= xrange[1]))[0]
                if (len(idx1) == 0 or len(idx2) == 0):
                    if (zrange[0] is None and
                            zrange[1] is None):
                        # this is for safety maybe not necessary
                        if scale == 'log':
                            z = mask_negative(z)
                        self._update_range(
                            zrange, [np.nanmin(z), np.nanmax(z)])
                        zrange = update_zrange(z)
                else:
                    zt = z[idx1, :]
                    zt = zt[:, idx2]
                    if scale == 'log':
                        zt = mask_negative(zt)
                    self._update_range(zrange, [np.nanmin(zt), np.nanmax(zt)])
            else:
                self._update_range(zrange, [np.nanmin(z), np.nanmax(z)])

        return zrange

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        cdata = self.getvar('cdata')
        cz = self.getvar('cz')
        if not cz:
            return crange
        if cdata is None:
            x, y, z = self._eval_xy()
            crange = self._update_range(crange,
                                        (np.nanmin(z), np.nanmax(z)))

        else:
            if np.iscomplexobj(cdata):
                tmp = np.max(np.abs(cdata))
                crange = self._update_range(crange,
                                            (-tmp, tmp))

            else:
                crange = self._update_range(crange,
                                            (np.nanmin(cdata), np.nanmax(cdata)))

        return crange

    def get_mappable(self):
        if self.getvar('cz'):
            return [a for a in self._artists if isinstance(a, ScalarMappable)]
        else:
            return []

    def handle_axes_change(self, data):
        name = data['name']
        if name.startswith('c') and self.getvar('cz'):
            self.del_artist(delall=True)
            self.delp('loaded_property')
            self.generate_artist()

    @classmethod
    def _saveload_names(self):
        return {'x', 'y', 'z'}

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        names = self._saveload_names()
        var = {name: check(self, name) for name in names}
        for name in names:
            if not var[name]:
                var[name+'data'] = self.getp(name)

        data['FigSurface'] = (1, var)
        data = super(FigSurface, self).save_data2(data)
        return data

    def load_data2(self, data):
        d = data['FigSurface']
        super(FigSurface, self).load_data2(data)
        var = d[1]

        names = self._saveload_names()
        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def _eval_xy(self):
        names = ("x", "y", "z")
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return [None]*len(names)
        return self.getp(names)

    def get_export_val(self, a):
        x, y, z = self.getvar("x", "y", "z")
        val = {"x": x, "y": y, "z": z}
        if self.getvar('cz'):
            val['cdata'] = self.getvar("cdata")
        return val

class FigRevolve(FigSurface):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._objs = []  # for debug....
            obj._data_extent = None
            obj._coarse_artist = None
            obj._fine_artist = None
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_var('r', ['iter|nonstr', 'dynamic'])
        p.add_var('z', ['iter|nonstr', 'dynamic'])

        p.add_key('cmap', None)
        p.add_key('shade', 'flat')
        p.add_key('linewidth', 1.0)
        p.add_key('alpha', 1)
        p.add_key('rstride', 1)
        p.add_key('cstride', 1)
        p.add_key('edgecolor', 'k')
        p.add_key('facecolor', 'b')
        p.add_key('cz',  False)
        p.add_key('cdata', None)

        p.set_ndconvert("x", "y")

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)

        obj = set_hidden_vars(obj)
        if v['cmap'] is not None:
            if isinstance(v['cmap'], Colormap):
                v['cmap'] = v['cmap'].name
            kywds['cmap'] = v['cmap']
            del v['cmap']

        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    def __init__(self, *args,  **kywds):
        FigSurface.__init__(self, *args,  **kywds)
        self._method = 'plot_revolve'

    @classmethod
    def get_namebase(self):
        return 'revolve'

    def _args2var(self):
        names0 = self.attr_in_file()
        names = ["r", "z"]
        use_np = [True]
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)
        r = values[0]
        z = values[1]
        if r is None or z is None:
            return False
        self.setp("r", r)
        self.setp("z", z)
        return True

    def generate_artist(self, coarse=False):
        if (self.isempty() is False
                and not coarse):
            return

        axes = self.get_figaxes()
        if not axes.get_3d():
            return

        container = self.get_container()
        r, z = self._eval_rz()
        # use_var should be false if evaluation is
        # okey.
        if self.getp('use_var'):
            return

        kywds = self._var["kywds"]

        cax = self.get_caxisparam()
        if cax is None:
            dprint1('Error: cax is None')
            return

        ### stride is handled in fig_surface
        kywds['cstride'] = 1
        kywds['rstride'] = 1
        if coarse:
            f = ifigure._visual_config["3d_rot_accel_factor"]
        else:
            f = 1

        if (not 'color' in kywds and
                not 'facecolors' in kywds):
            cmap = self.get_cmap()
            kywds['cmap'] = get_cmap(cmap)

        kywds['shade'] = self.getvar('shade')
        # if self.getp('alpha') is not None else 1
        kywds['alpha'] = self.getp('alpha')
        fc = self.getp('facecolor')
        if isinstance(fc, str):
            fc = cc.to_rgba(fc)
        if fc is None:
            fc = [0, 0, 0, 0]
        else:
            fc = list(fc)
            if self.getp('alpha') is not None:
                fc[3] = self.getp('alpha')
        ec = self.getp('edgecolor')
        if isinstance(ec, str):
            ec = cc.to_rgba(ec)
        if ec is None:
            ec = [0, 0, 0, 0]
        else:
            ec = list(ec)
            if self.getp('alpha') is not None:
                ec[3] = self.getp('alpha')
        cz = self.getvar('cz')
        if cz is None:
            kywds['cz'] = False
        else:
            kywds['cz'] = cz
        if kywds['cz']:
            kywds['cdata'] = self.getvar('cdata')
        kywds['facecolor'] = (fc,)
        kywds['edgecolor'] = (ec,)
        kywds['linewidths'] = 0.0 if self.getp(
            'linewidth') is None else self.getp('linewidth')

        m = getattr(container, self._method)
        self._artists = [m(r, z, **kywds)]
        self._fine_artist = self._artists[0]

        for artist in self._artists:
            artist.do_stencil_test = False
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))
            self._objs.append(weakref.ref(artist))
            cax.set_crangeparam_to_artist(artist)

    def get_data_extent(self):
        if self._data_extent is not None:
            return self._data_extent
        r, z = self._eval_rz()
        self._data_extent = [-np.max(r), np.max(r),
                             -np.max(r), np.max(r),
                             np.min(z), np.max(z)]
        return self._data_extent

    def get_xrange(self, xrange=[None, None], scale='linear'):
        r, z = self._eval_rz()
        if r is None:
            return xrange
        xrange = self._update_range_gl(xrange, 0)
        if scale == 'log':
            return self._update_range(xrange, [1e-15, np.nanmax(xrange)])
        return xrange

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        r, z = self._eval_rz()
        if r is None:
            return yrange
        yrange = self._update_range_gl(yrange, 1)
        if scale == 'log':
            return self._update_range(yrange, [1e-15, np.nanmax(yrange)])
        return yrange

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        r, z = self._eval_rz()
        zrange = self._update_range_gl(zrange, 2)
        if scale == 'log':
            return self._update_range(zrange, [1e-15, np.nanmax(zrange)])
        return zrange

    @classmethod
    def _saveload_names(self):
        return {'r', 'z'}

    def _eval_rz(self):
        names = ("r", "z")
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return [None]*len(names)
        return self.getp(names)

    def get_shade(self, a=None):
        return self.getvar('shade')

    def get_export_val(self, a):
        r, z = self.getvar("r", "z",)
        val = {"r": r, "z": z}
        return val



from .file_helper_dialog import get_handler
import os
from ifigure.utils.add_sys_path import AddSysPath
from ifigure.widgets.dlg_preference import PrefComponent
from ifigure.utils.setting_parser import iFigureSettingParser as SP
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('HelperApp')


class AdvancedConfig(PrefComponent):
    def __init__(self):
        PrefComponent.__init__(self, 'Advanced')
        p = SP()
        self.setting = p.read_setting('pref.advanced_config')

    def save_setting(self):
        p = SP()
        p.write_setting('pref.advanced_config', self.setting)

    def get_dialoglist(self):
        path_txt = '\n'.join(self.setting["user_path"].split(':'))
        list1 = [["Max worker threads number", str(self.setting["max_thread"]), 0, None],
                 ["Text in vector format file", not self.setting["keep_text_as_text"], 3, {
                     "text": "Rasterize Text"}],
                 ["dpi of image in print", str(
                     self.setting["image_dpi"]), 0, None],
                 ["default repository", str(
                     self.setting["hg_default_url"]), 0, None],
                 ["user path", path_txt, 235, {"nlines": 4}],
                 ["file actions",  None, 41,
                  {"label": "Edit...", 'func': get_handler, 'noexpand': True}], ]

        hint1 = ['Maximum number of threads used to run modeling scripts simultaneously',
                 'Non-rasterized text is generated by LaTeX, which may result in longer processing time to generate a PDF/EPS file. Also, text may look differently from what is seen on your screen (requires ghostscript, poppler/xpdf, latex installations)',
                 'dpi (dot per inch) for raster image in EPS/PDF files',
                 'default Mercurial repository',
                 'additinal python path',
                 'edit actions when opening a file']

        return list1, hint1

    def set_dialog_result(self, value):
        self.setting["max_thread"] = int(value[0])
        self.setting["keep_text_as_text"] = not value[1]
        self.setting["image_dpi"] = int(value[2])
        self.setting["hg_default_url"] = str(value[3])
        self.setting["user_path"] = ':'.join(value[4].split('\n'))
        self.add_user_path()

    def add_user_path(self):
        for item in self.setting['user_path'].split(':'):
            AddSysPath(os.path.expanduser(item))
